import os
import sys
import itertools
import webbrowser
import logging
import urllib
import base64
import json
import time

import c4d

try:
    import urllib.request as urllib2
except ImportError:
    import urllib2

PY3 = sys.version_info[0] == 3

if PY3:
    lib_path = os.path.join(os.path.dirname(__file__),'utf-8')
else:
    lib_path = os.path.join(unicode(os.path.dirname(__file__),'utf-8'))
sys.path.insert(0, lib_path)


logger = logging.getLogger('cvMenus')

CVMENU_ID = 1032060
ABOUTDLG_ID = 1032059
BUGDLG_ID = 1032058

class CVMenu():

    def __init__(self, dlg, name):
        self._menus = {}
        self._dlg = dlg
        self._name = name
        self._libpath = os.path.dirname(os.path.abspath(__file__))
        print(os.path.basename(self._libpath))
        if os.path.basename(os.path.dirname(self._libpath)) == 'lib':
            self._libpath = os.path.dirname(self._libpath)
        self._respath = os.path.join(os.path.dirname(self._libpath),'res')
        self._ids = itertools.count(1000)

    def GenerateID(self):
        """
        Generate & return a new ID
        """
        return next(self._ids)

    def Add(self, id=None, name='', help='', command=None, *args, **kwargs):
        """
        Add a new menu item
        """
        if not id:
            id = self.GenerateID()
        if help == '':
            help = name
        self._menus[id]={'name':name, 'help':help, 'command':command, 'args':args, 'kwargs':kwargs}
        return id

    def AddSeparator(self):
        self.Add(0,'-')

    def SetName(self, id, name):
        """
        Set the name of the menu item with specific ID
        """
        self._menus[id]['name'] = name

    def SetHelp(self, id, help):
        """
        Set the help text of the menu item with specific ID
        """
        self._menus[id]['help'] = help

    def SetCommand(self, id, func, *args, **kwargs):
        """
        Set the command of the menu item with specific ID
        """
        self._menus[id]['command'] = func
        self._menus[id]['args'] = args
        self._menus[id]['kwargs'] = kwargs

    def Find(self,name):
        """
        Find & return the ID of the first menu item with the specified name
        """
        for k,v in list(self._menus.items()):
            if v['name'] == name:
                return k

    def CreateMenu(self):
        """
        Create the menu within the GeDialog
        """
        self._dlg.MenuSubBegin(self._name)
        for k,v in list(self._menus.items()):
            if v['name'] == '-':
                self._dlg.MenuAddSeparator()
            else:
                self._dlg.MenuAddString(k,v['name'])
        self._dlg.MenuSubEnd()

    def Command(self, id):
        """
        Execute the command associated with a specified menu ID
        """
        def wrapper(commandDict):
            try:
                func = commandDict['command']
                args = commandDict['args']
                kwargs = commandDict['kwargs']
                return func(*args,**kwargs)
            except Exception as e:
                logger.exception("Command id %s exception", id)
        if id in self._menus:
            logger.debug("Executing command id %s", id)
            wrapper(self._menus[id])
        return True

class HelpMenu(CVMenu):

    def __init__(self, dlg, name, ver, credit='', image=None):
        CVMenu.__init__(self,dlg,'Help')

        self._aboutDlg = None
        self._bugDlg = None

        self._pluginVer = ver

        self.bitbucket = {}
        self.mediafire = {}

        self.reportBugs = False
        self.bugAttachments = []

        # Build the menu
        self._helpID = self.Add(name='Show Help...')
        self._tutorialID = self.Add(name='Show Tutorials...')
        self._supportID = self.Add(name='Support...')
        self.AddSeparator()
        self._aboutID = self.Add(name='About...')
        if self.reportBugs:
            self.AddSeparator()
            self._reportID = self.Add(name='Report Bug/Idea...')
            self.SetCommand(self._reportID,self.OpenBugDialog)

        self.SetCommand(self._aboutID,self.OpenAboutDialog,name,ver,credit,image)

    def SetHelpURL(self,url):
        self.SetCommand(self._helpID,webbrowser.open,url)

    def SetTutorialURL(self,url):
        self.SetCommand(self._tutorialID,webbrowser.open,url)

    def SetSupportURL(self,url):
        self.SetCommand(self._supportID,webbrowser.open,url)

    def AddBugAttachment(self,fn):
        self.bugAttachments.append(fn)

    def OpenAboutDialog(self, name, ver, credit='', image=''):

        self._aboutDlg = AboutDialog()
        self._aboutDlg.name = name
        self._aboutDlg.ver = ver
        self._aboutDlg.credit = credit
        self._aboutDlg.splashImage = os.path.join(self._respath,image)
        logger.debug("Splash Image: %s",self._aboutDlg.splashImage)

        #Determine where the user's cursor is
        state = c4d.BaseContainer()
        c4d.gui.GetInputState(c4d.BFM_INPUT_MOUSE, c4d.MOUSE_MOVE, state)
        pos_x = state.GetInt32(c4d.BFM_INPUT_X)
        pos_y = state.GetInt32(c4d.BFM_INPUT_Y)

        #Get the screen dimensions for the monitor where the cursor resides
        dimensions = c4d.gui.GeGetScreenDimensions(x=pos_x, y=pos_y, whole_screen=False)
        size_x = dimensions["sx2"] - dimensions["sx1"]
        size_y = dimensions["sy2"] - dimensions["sy1"]

        splash_w, splash_h = self._aboutDlg.InitSplashBmp()
        logger.debug("Splash Size: %s x %s", splash_w, splash_h)

        #Open the about dialog
        self._aboutDlg.Open(dlgtype=c4d.DLG_TYPE_ASYNC_POPUPEDIT, subid=0,
                            xpos=size_x/2 - splash_w/2, ypos=size_y/2 - splash_h/2,
                            defaultw=splash_w, defaulth=splash_h)

        return True

    def OpenBugDialog(self):
        if self._bugDlg is None:
            self._bugDlg = BugDialog(self.bitbucket,self.mediafire,self._pluginVer)
            self._bugDlg.attachments = self.bugAttachments
        self._bugDlg.Open(pluginid=CVMENU_ID, subid=BUGDLG_ID,
                          dlgtype=c4d.DLG_TYPE_ASYNC,
                          defaultw=320, defaulth=480)

class AboutDialog(c4d.gui.GeDialog):
    def __init__(self):
        self.name = ''
        self.ver = ''
        self.credit = ''
        self.splashImage = 'splash.png'
        self.splashBmp = None
        self.splashBmpText = None
        self.splashBtn = None
        self._splashWidth = 0
        self._splashHeight = 0

    def InitSplashBmp(self):
        self.splashBmp = c4d.bitmaps.BaseBitmap()
        result,movie = self.splashBmp.InitWith(self.splashImage)
        if result == c4d.IMAGERESULT_OK:
            self._splashWidth = self.splashBmp.GetBw()
            self._splashHeight = self.splashBmp.GetBh()
            return self._splashWidth, self._splashHeight
        else:
            logger.error("Failed to load splash bitmap %s [%s]",self.splashImage,result)
            return 0, 0

    def CreateLayout(self):
        #Load the Banner Graphic
        if not self.splashBmp:
            self.InitSplashBmp()

        #Prepare to add text to Banner Graphic
        self.splashBmpText = c4d.bitmaps.GeClipMap()
        self.splashBmpText.InitWithBitmap(self.splashBmp)

        #Start adding text
        self.splashBmpText.BeginDraw()

        #Calculate Version Number size/placement
        version_text = "%s v%s" % (self.name, self.ver)

        #Calculate text sizes
        text_w = self.splashBmpText.TextWidth(version_text)
        text_h = self.splashBmpText.TextHeight()

        #Add Version Number
        offset_x = -32
        offset_y = -5
        self.splashBmpText.TextAt(self._splashWidth - text_w + offset_x, self._splashHeight - text_h + offset_y, version_text)

        #Add the About Information
        credit_block = self.credit.split("\n")

        #Hardcoded based on banner graphic
        ### TODO: Add method for placing credit area
        cursor_x = 282
        cursor_y = 79 + text_h
        block_width = 202

        for creditLine in credit_block:
            text_w = self.splashBmpText.TextWidth(creditLine)
            self.splashBmpText.TextAt(cursor_x + block_width - text_w, cursor_y, creditLine)
            cursor_y += text_h

        #Finish adding text to banner graphic
        self.splashBmpText.EndDraw()

        #Create the Dialog Elements
        self.SetTitle("About " + self.name)
        self.splashBtn = self.AddCustomGui(id=1000, pluginid=c4d.CUSTOMGUI_BITMAPBUTTON,
                                           name='About', flags=c4d.BFH_FIT | c4d.BFH_FIT,
                                           minw=5, minh=5)

        self.splashBtn.SetImage(self.splashBmpText.GetBitmap())

        return True

    def Command(self, id, msg):
        if id == 1000:
           self.Close()
        return True

class StatusUserArea(c4d.gui.GeUserArea):

    def __init__(self):
        self._box = 0
        self.pct = 0

    def DrawUnfilledRect(self,x1,y1,x2,y2):
        self.DrawLine(x1,y1,x2,y1)
        self.DrawLine(x1,y1,x1,y2)
        self.DrawLine(x1,y2,x2,y2)
        self.DrawLine(x2,y1,x2,y2)

    def DrawMsg(self,x1,y1,x2,y2,msg):

        self.DrawSetPen(c4d.COLOR_BG)
        self.DrawRectangle(x1,y1,x2,y2)

        #self.OffScreenOn()
        self.SetClippingRegion(x1,y1,x2,y2)

        # Set Colors
        fg_color = c4d.COLOR_PROGRESSBAR
        pt_color = c4d.COLOR_TEXT_SELECTED
        bg_color = c4d.COLOR_BG

        # Set Box Size
        box_size = 15
        margin = 10

        mod = self._box % 10
        if mod > 5:
            mod = 10-mod

        for i in range(5):
            x = (((x2-y2)/5)*i) + 20

            # Draw box, filled or unfilled
            if i == mod-1:
                self.DrawSetPen(fg_color)
            else:
                self.DrawSetPen(bg_color)
            self.DrawRectangle(x,margin,x+box_size,margin+box_size)

            # Draw outline, indicating completion
            if self.pct >= i/5.0:
                self.DrawSetPen(pt_color)
            else:
                self.DrawSetPen(fg_color)
            self.DrawUnfilledRect(x,margin,x+box_size,margin+box_size)

    def Timer(self,msg):
        self._box += 1
        self.Redraw()

class BugDialog(c4d.gui.GeDialog):

    statusArea = StatusUserArea()

    ids = type('', (),{
       'USERNAME': 1000,
       'USEREMAIL': 1001,
       'BUGTYPE': 1002,
       'BUGTITLE': 1003,
       'BUGDESC': 1004,
       'BUGFILES': 1005,
       'BUGFILES_ADD': 1006,
       'SCREEN': 2000,
       'TAB_ENTRY': 2001,
       'TAB_STATUS': 2002,
       'TAB_FINISHED': 2003,
       'UA_STATUS': 3000,
       'TXT_STATUS': 3001,
       'TXT_ERR': 3002,
       'MSG_SUBMIT_STATUS': 1031912,
    })

    def __init__(self,bitbucket={},mediafire={},ver=""):
        self.allowAttachments = False
        self.attachments = []
        self.bugThread = None
        self.bitbucket = bitbucket
        self.mediafire = mediafire
        self._pluginVer = ver

    def CreateLayout(self):
        # Create dialog elements
        self.SetTitle('Report Bug/Idea')
        self.GroupBegin(0,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,1)
        self.GroupSpace(10,10)
        self.GroupBorderSpace(5,5,5,5)

        self.TabGroupBegin(self.ids.SCREEN,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,c4d.TAB_NOSELECT)

        self.GroupBegin(self.ids.TAB_ENTRY,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,1)

        self.AddStaticText(self.ids.TXT_ERR,c4d.BFH_CENTER|c4d.BFV_CENTER,300,10,'')

        # 2 column section
        self.GroupBegin(0,c4d.BFH_SCALEFIT|c4d.BFV_TOP,2)
        self.AddStaticText(0,c4d.BFH_LEFT|c4d.BFV_CENTER,name='Your Name')
        self.AddEditText(self.ids.USERNAME,c4d.BFH_SCALEFIT|c4d.BFV_CENTER)
        self.AddStaticText(0,c4d.BFH_LEFT|c4d.BFV_CENTER,name='Your Email')
        self.AddEditText(self.ids.USEREMAIL,c4d.BFH_SCALEFIT|c4d.BFV_CENTER)
        self.AddStaticText(0,c4d.BFH_LEFT|c4d.BFV_CENTER,name='Issue Type')
        self.AddComboBox(self.ids.BUGTYPE,c4d.BFH_SCALEFIT|c4d.BFV_CENTER)
        self.AddChild(self.ids.BUGTYPE,0,'Bug')
        self.AddChild(self.ids.BUGTYPE,1,'Idea')
        self.AddStaticText(0,c4d.BFH_LEFT|c4d.BFV_CENTER,name='Issue Title')
        self.AddEditText(self.ids.BUGTITLE,c4d.BFH_SCALEFIT|c4d.BFV_CENTER)
        self.GroupEnd()

        # 1 column section
        self.GroupBegin(0,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,1)
        self.AddStaticText(0,c4d.BFH_LEFT|c4d.BFV_CENTER,name='Description')
        self.AddMultiLineEditText(self.ids.BUGDESC,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,style=c4d.DR_MULTILINE_WORDWRAP)
        self.GroupEnd()

        if self.allowAttachments:
            # Files section
            self.GroupBegin(0,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,1)
            self.GroupBegin(0,c4d.BFH_SCALEFIT|c4d.BFV_TOP,2)
            self.AddStaticText(0,c4d.BFH_SCALEFIT|c4d.BFV_CENTER,name='Files (drag onto dialog)')
            self.AddButton(self.ids.BUGFILES_ADD,c4d.BFH_RIGHT|c4d.BFV_CENTER,30,10,'Add...')
            self.GroupEnd()
            self.AddMultiLineEditText(self.ids.BUGFILES,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,style=c4d.DR_MULTILINE_READONLY)
            self.GroupEnd()

        # Bottom buttons
        self.GroupBegin(0,c4d.BFH_SCALEFIT|c4d.BFV_BOTTOM,1)
        self.AddDlgGroup(c4d.DLG_OK|c4d.DLG_CANCEL)
        self.GroupEnd()

        self.GroupEnd()

        # Status Screen
        self.GroupBegin(self.ids.TAB_STATUS,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,1)
        self.GroupBegin(0,c4d.BFH_SCALE|c4d.BFV_SCALE|c4d.BFH_CENTER|c4d.BFV_CENTER,1)
        self.AddStaticText(0,c4d.BFH_SCALE|c4d.BFV_SCALE|c4d.BFH_CENTER|c4d.BFV_CENTER,300,30,'Please Wait. Submitting Bug')
        self.AddUserArea(self.ids.UA_STATUS,c4d.BFH_SCALE|c4d.BFV_SCALE|c4d.BFH_CENTER|c4d.BFV_CENTER,300,40)
        self.AttachUserArea(self.statusArea,self.ids.UA_STATUS)
        self.statusArea.SetTimer(1000)
        self.AddStaticText(self.ids.TXT_STATUS,c4d.BFH_SCALE|c4d.BFV_SCALE|c4d.BFH_CENTER|c4d.BFV_CENTER,300,30,'')
        self.GroupEnd()
        self.GroupEnd()

        # Finished Screen
        self.GroupBegin(self.ids.TAB_FINISHED,c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,1)
        self.GroupBegin(0,c4d.BFH_SCALE|c4d.BFV_SCALE|c4d.BFH_CENTER|c4d.BFV_CENTER,1)
        self.AddStaticText(0,c4d.BFH_SCALE|c4d.BFV_SCALE|c4d.BFH_CENTER|c4d.BFV_CENTER,300,30,'Bug Submitted Successfully.')
        self.AddStaticText(0,c4d.BFH_SCALE|c4d.BFV_SCALE|c4d.BFH_CENTER|c4d.BFV_CENTER,300,30,'Thanks for your feedback.')
        self.AddButton(c4d.DLG_CANCEL,c4d.BFH_SCALE|c4d.BFV_SCALE|c4d.BFH_CENTER|c4d.BFV_CENTER,200,20,'Close')
        self.GroupEnd()
        self.GroupEnd()

        self.GroupEnd()

        self.GroupEnd()
        return True

    def InitValues(self):
        if c4d.GetC4DVersion() < 20500:
            si = c4d.GeGetSerialInfo(c4d.SERIALINFO_CINEMA4D)
        else:
            si = c4d.GeGetSerialInfo()
        if si and 'name' in si:
            self.SetString(self.ids.USERNAME,si['name'])

        authBc = c4d.plugins.GetWorldPluginData(1031621)
        #for k,v in authBc:
        #    print k,"=",v
        if authBc:
            self.SetString(self.ids.USEREMAIL,authBc.GetString(20002))

        if self.attachments:
            self.SetString(self.ids.BUGFILES,'\n'.join(self.attachments))

        return True

    def Command(self,id,msg):
        if id==c4d.DLG_CANCEL:
            self.Close()

        elif id==c4d.DLG_OK:
            self.bugThread = BugSubmitThread(self.bitbucket,self.mediafire,self.ids.MSG_SUBMIT_STATUS)
            self.bugThread.userName = self.GetString(self.ids.USERNAME)
            self.bugThread.userEmail = self.GetString(self.ids.USEREMAIL)
            bugTypeInt = self.GetLong(self.ids.BUGTYPE)
            if bugTypeInt == 0:
                self.bugThread.bugType = 'bug'
            else:
                self.bugThread.bugType = 'proposal'
            self.bugThread.bugTitle = self.GetString(self.ids.BUGTITLE)
            self.bugThread.bugDesc = self.GetString(self.ids.BUGDESC)
            self.bugThread.bugFiles = self.GetString(self.ids.BUGFILES).split('\n')
            self.bugThread.pluginVer = self._pluginVer
            self.SetLong(self.ids.SCREEN,self.ids.TAB_STATUS)
            c4d.StatusSetSpin()
            self.bugThread.Start()

        elif id==self.ids.BUGFILES_ADD:
            fn = c4d.storage.LoadDialog()
            if fn:
                self.attachments.append(fn)
                self.SetString(self.ids.BUGFILES,'\n'.join(self.attachments))

        return True

    def Message(self,msg,result):
        if msg.GetId() == c4d.BFM_DRAGRECEIVE:
            dragDict = self.GetDragObject(msg)

            if dragDict['type'] == c4d.DRAGTYPE_FILES or \
               dragDict['type'] == c4d.DRAGTYPE_FILENAME_IMAGE or \
               dragDict['type'] == c4d.DRAGTYPE_FILENAME_SCENE or \
               dragDict['type'] == c4d.DRAGTYPE_FILENAME_OTHER:

                if msg[c4d.BFM_DRAG_FINISHED]:
                    self.attachments.append(dragDict['object'])
                    self.SetString(self.ids.BUGFILES,'\n'.join(self.attachments))

                return self.SetDragDestination(c4d.MOUSE_INSERTCOPY,self.ids.BUGFILES)


        return c4d.gui.GeDialog.Message(self,msg,result)

    def CoreMessage(self,id,msg):
        if id == self.ids.MSG_SUBMIT_STATUS:
            status = self.bugThread.status
            statusTxt = self.bugThread.statusTxt
            if status == 1:
                self.SetLong(self.ids.SCREEN,self.ids.TAB_FINISHED)
                c4d.StatusClear()
            elif status == -2:
                self.SetDefaultColor(self.ids.TXT_ERR, c4d.COLOR_TEXT, c4d.Vector(1,0,0))
                self.SetString(self.ids.TXT_ERR,"ERROR: " + statusTxt)
                self.SetLong(self.ids.SCREEN,self.ids.TAB_ENTRY)
            else:
                c4d.StatusSetSpin()
                c4d.StatusSetText(statusTxt)
                self.statusArea.pct = status
                self.SetString(self.ids.TXT_STATUS,statusTxt)
        return True

class BugSubmitThread(c4d.threading.C4DThread):

    def __init__(self,bitbucket=None,mediafire=None,statusID=None):

        self.bitbucketAccount = ''
        self.bitbucketRepo = ''
        self.bitbucketUser = ''
        self.bitbucketPass = ''

        self.mediafireUser = ''
        self.mediafirePass = ''
        self.mediafireAppID = ''
        self.mediafireApiKey = ''

        try:
            self.bitbucketAccount = bitbucket['account']
            self.bitbucketRepo = bitbucket['repo']
            self.bitbucketUser = bitbucket['user']
            self.bitbucketPass = bitbucket['pass']
            self.mediafireUser = mediafire['user']
            self.mediafirePass = mediafire['pass']
            self.mediafireAppID = mediafire['appid']
            self.mediafireApiKey = mediafire['apikey']
        except KeyError:
            pass

        self.userName = ''
        self.userEmail = ''
        self.bugType = ''
        self.bugTitle = ''
        self.bugDesc = ''
        self.bugFiles = []
        self.bugLinks = []
        self.pluginVer = ''

        self.statusID = statusID
        self.status = -1
        self.statusTxt = ''

    def Main(self):
        self.bugFiles = [i for i in self.bugFiles if i != '']
        if self.bugFiles and self.mediafireUser and self.mediafirePass and self.mediafireAppID and self.mediafireApiKey:
            #Disable mediafire upload for now
            #self.UploadFiles()
            pass
        self.UpdateStatus(0.8,'Submitting Report Text')
        id = self.CreateBitIssue()
        if id:
            self.UpdateStatus(1)
        else:
            self.UpdateStatus(-2,'Bug Submit Failed. Please try again.')

    def UpdateStatus(self,status=-1,statusTxt=''):
        self.status = status
        self.statusTxt = statusTxt
        if self.statusID:
            c4d.SpecialEventAdd(self.statusID)

    def CreateBitIssue(self):
        url = 'https://bitbucket.org/api/1.0/repositories/'+self.bitbucketAccount+'/'+self.bitbucketRepo+'/issues/'
        req = urllib2.Request(url)

        content = 'Reported By: %s (%s)\n\n' % (self.userName,self.userEmail)
        content += 'Reported with Version: %s\n\n\n' % self.pluginVer
        content += self.bugDesc
        if self.bugLinks:
            content += '\n\nFiles:\n' + '\n'.join(self.bugLinks)

        req.data = urllib.parse.urlencode({'title':self.bugTitle,
                                        'kind':self.bugType,
                                        'status':'new',
                                        'pluginVer':str(self.pluginVer),
                                        'content':content
                                        }).encode('utf-8')
        logger.debug("BitBucket Request: %s",req.data)
        authStr = "%s:%s" % (self.bitbucketUser,self.bitbucketPass)
        auth = base64.urlsafe_b64encode(authStr.encode('utf-8'))
        req.add_header('Authorization', ' Basic %s' % auth)

        try:
            response = urllib2.urlopen(req)
            html = response.read()
            bitResponse = json.loads(html)
            logger.debug("BitBucket Response: %s",bitResponse)
            return bitResponse['local_id']
        except urllib.error.URLError as e:
            self.UpdateStatus(-2,'Bug Submit Failed')
            logger.exception("Bug Submit Failed: %s" % e)

        return False

    def UploadFiles(self):
        mf = mediafire.MediaFireLib(self.mediafireUser,self.mediafirePass,self.mediafireAppID,self.mediafireApiKey)

        try:
            tok = mf.user_getSessionToken()
        except urllib.error.URLError as e:
            logger.exception("Error getting mediafire session token: %s", e)
            return False

        self.UpdateStatus(.2,'Preparing to upload files')

        try:
            cnt = len(self.bugFiles)

            # Create Folder for Bug
            folderName = 'bug_' + self.bitbucketRepo + str(int(time.time()))
            folder = mf.folder_create(folderName)
            if folder['result'] != 'Success':
                logger.error("Folder creation failed %s", folder)
                return False
            folderKey = folder['folder_key']
            self.UpdateStatus(.3,'Uploading file 1 of %s' % cnt)

            logger.debug("Attached files: %s", self.bugFiles)
            for i,f in enumerate(self.bugFiles):
                #f = unicode(f,'utf-8')
                if not os.path.exists(f):
                    continue

                fileKey = mf.upload_UploadFile(f,folderKey)
                logger.debug("Upload File Response: %s", fileKey)

                while True:
                    progress = mf.Upload_PollUpload(fileKey)
                    logger.debug("Poll Upload Response: %s", progress)
                    if progress['status'] == '99':
                        fileQKey = progress['quickkey']
                        break
                    self.UpdateStatus(.4+(float(i)/cnt*.2),'Uploading file %s of %s' % (i+1,cnt))
                    time.sleep(3)

                fileLinkResponse = mf.file_getLinks(fileQKey)
                logger.debug("File Link Response: %s", fileLinkResponse)
                if 'links' in fileLinkResponse:
                    for link in fileLinkResponse['links']:
                        self.bugLinks.append(link['direct_download'])
                self.UpdateStatus(.4+(float(i)/cnt*.4),'File %s of %s Uploaded Successfully' % (i+1,cnt))

        except urllib.error.URLError as e:
            self.UpdateStatus(-2,'Upload failed. Upload error.')
            logger.exception("Mediafire API Error: %s", e)
            return False

        except IOError as e:
            self.UpdateStatus(-2,'Upload failed. File not found.')
            logger.exception("File Error: %s", e)
            return False