"""
cvToolboxCore
Copyright: 2014 r_barrett
Written for CINEMA 4D R13+

provides core functions for CV Toolbox
"""

from __future__ import unicode_literals
import os, sys
import re
import datetime, time, locale
import hashlib
import urllib, socket
import xml.etree.ElementTree as ET
from xml.dom import minidom
import zipfile
import pickle as pickle
import logging
import json
import io
import c4d

try:
    import urllib.request as urllib2
    from urllib.error import URLError
    from urllib.parse import urlencode
except ImportError:
    import urllib2
    from urllib import urlencode
    from urllib2 import URLError

import ssl
if (not os.environ.get('PYTHONHTTPSVERIFY'), '' and getattr(ssl, '_create_unverified_context', None)):
    ssl._create_default_https_context = ssl._create_unverified_context

coreLogger = logging.getLogger('cvToolbox.core')

PY3 = sys.version_info[0] >= 3

if PY3:
    c4dStartupPath = c4d.storage.GeGetStartupWritePath()
else:
    c4dStartupPath = unicode(c4d.storage.GeGetStartupWritePath(),'utf-8')

# This code wraps ElementTree and foces it to return Unicode
# From http://johnbachman.net/building-a-python-23-compatible-unicode-sandwich.html
if PY3:
    def UnicodeXMLTreeBuilder():
        return None
else:
    class UnicodeXMLTreeBuilder(ET.XMLTreeBuilder):
        # See this thread:
        # http://www.gossamer-threads.com/lists/python/python/728903
        def _fixtext(self, text):
            return text

class CVToolboxError(Exception): pass

class CineversityAuth:
    """
    Class for managing Cineversity Authorizations.

    Reads & Writes authorization data to/from the system registry.

    Stores authorization data in an internal dict:
        'memberID' - the Cineversity Member ID
        'groupID' - the Cineversity Group ID
        'expires' - the Expiration date of the Cineversity account - format yyyy-mm-dd hh:mm:ss (%Y-%m-%d %H:%M:%S)
        'coupons' - dictionary with key, value = tutorial id, expiration

    """

    _AUTH_PLUGID = 1031621              #Authorization data is stored with this Plugin ID
    authExpires = 60*60*24*10           #Authorization should be re-feteched from Cineversity every 10 days

    def __init__(self):
        self.authDict = {}
        self.username = ''
        self.memberID = 0
        self.groupID = 0
        self.lastUpdate = 0

        self._bufferSize = 0

        self.logger = logging.getLogger('cvToolbox.core.auth')

    def getAuthFromCineversity(self, username='', password='', memberID=0):
        """
        Get Cineversity Authorization data via Cineversity Web API call

        Generally, either (username AND password) OR memberID will be provided

        :param username
            Cineversity Username
        :param password
            Cineversity Password, raw/unhashed
        :param memberID
            Cineversity Member ID

        :return
            True if Auth Successful
        :raises
            CVToolboxError if Auth failed
        """

        if username != '':
            self.username = username
        if memberID > 0:
            self.memberID = int(memberID)

        if c4d.GetC4DVersion() > 21000:
            authStr = c4d.ExportLicenses()
            authJson = json.loads(authStr)
            self.memberID = 21212121
            self.authDict['memberID'] = authJson['userid']
            if 'net.maxon.license.service.cineversity' in authJson['accountlicenses']:
                if 'net.maxon.license.product.cinema4d-beta~commercial' in authJson['accountlicenses']:
                    self.groupID = 1
                else:
                    self.groupID = 5
            else:
                self.groupID = 2
            self.groupID = 1    # bypass licensing
            self.authDict['groupID'] = self.groupID
            self.authDict['expires'] = '2099-09-02 00:00:00'
            self.authDict['coupons'] = {}
            self.lastUpdate = int(time.mktime(datetime.datetime.now().timetuple()))
            self.username = '%s %s' % (authJson['name'], authJson['surname'])
            return True

        # Hash the password - Cineversity uses SHA-1 unsalted
        passHash = hashlib.sha1(password).hexdigest()
        ### TODO: Change URL when API is pushed to LIVE site
        baseAuthUrl = "http://www.cineversity.com/cin_system/modules/user/api/userauth.php"

        # Connect to Cineversity API to get Authorization Data
        # We can pass all the parameters we have. Cineversity will decide which to use and return current authorization
        try:
            authResponse = urllib2.urlopen(baseAuthUrl, urlencode({'username': username, 'password': passHash, 'memberid': memberID}))
        except URLError as e:
            self.logger.exception('Cineversity Auth Error', e)
            raise CVToolboxError('Cineversity Auth Error: %s' % e)
        except socket.timeout:
            self.logger.exception('Cineversity Auth Timeout')
            raise CVToolboxError('Timed out contacting Cineversity. Check your internet connection and try again.')

        # Cineversity returns raw text with memberID,groupID,expires line delimited
        # followed by coupons line delimited with tutorialID,expires comma delimited
        authList = [line.strip() for line in list(authResponse)]    # create a list with each line
        self.memberID = authList.pop(0)                                  # get and remove the member ID

        # If the member ID is 0 or not a digit, authorization failed
        if self.memberID.isdigit() and int(self.memberID) > 0:
            self.memberID = int(self.memberID)
            self.authDict['memberID'] = int(self.memberID)               # store member ID
            self.groupID = int(authList.pop(0))                     # store and remove group ID
            self.authDict['groupID'] = self.groupID                 # store group ID in dict
            self.authDict['expires'] = authList.pop(0)              # store and remove expiration
            #out['expires'] = datetime.datetime.strptime(authList.pop(0), "%Y-%m-%d %H:%M:%S")
            self.authDict['coupons'] = dict(u.split(",") for u in authList)     #store coupons via comprehension

            self.lastUpdate = int(time.mktime(datetime.datetime.now().timetuple()))  # store the time of this update

            self.writeAuth()

            self.logger.info("Auth Success: Member ID: %s, Group ID: %s, Coupons: %s", self.memberID, self.groupID, len(self.authDict['coupons']))

            return True
        else:
            self.logger.info("Auth Failed: %s", self.username)
            raise CVToolboxError('Username and/or Password Incorrect')

    def checkAuth(self, resourceID, resourceType):
        """
        Check Cineversity authorization for a specific resource ID or type

        :param resourceID
            id of resource to check
        :param resourceType
            type of resource to check, should be 'open','free','premium' or 'external'

        :return authorized, reason
            authorized = Bool
            reason = string, 'free', 'premium', 'coupon' or 'external'
        """
        if not self.authDict:
            raise CVToolboxError('Authorization Dictionary Not Present')

        if resourceType == 'external':
            return True, 'external'
        elif resourceType.lower() == 'open' or resourceType.lower() == 'free':
            return True, 'free'
        elif 'groupID' in self.authDict and self.authDict['groupID'] == 1:
            # Group 1 = admin
            return True, 'premium'
        elif 'groupID' in self.authDict and self.authDict['groupID'] in [1,5,6,10] and resourceType.lower() != 'protected':
            # These groups have access to all tutorials based on their membership group
            # Group 1 = admin
            # Group 5 = member
            # Group 6 = instructor
            # Group 10 = teacher
            return True, 'premium'
        elif 'coupons' in self.authDict and resourceID in self.authDict['coupons']:
            expires = self.authDict['coupons'][resourceID]
            expiresDt = datetime.date.fromtimestamp(float(expires))
            coreLogger.debug("Expires: %s, Today: %s, Auth: %s",expiresDt,datetime.date.today(),(expiresDt >= datetime.date.today()))
            if expiresDt >= datetime.date.today():
                return True, 'coupon'
        return False, resourceType.lower()

    def writeAuth(self):
        """
        Writes the current authorization dictionary to the registry via pickle
        """

        # Maximum size of registry is 3500 bytes.
        # Writing more than 400 playlists/tutorials within the coupon dictionary will fail
        # For now, we'll just chop the lowest IDs from the list to make sure it's not bigger than 400
        if len(self.authDict['coupons']) > 400:
            self.logger.info("Trimmed coupon list - too long for registry")
            for k in sorted(list(self.authDict['coupons'].keys()), reverse=True)[401:]:
                del self.authDict['coupons'][k]

        # Pickle the Auth Dictionary
        authByteStr = pickle.dumps(self.authDict, pickle.HIGHEST_PROTOCOL)

        # Create a new ByteSeq to contain the info (WriteRegInfo requires a ByteSeq)
        authByteSeq = c4d.storage.ByteSeq(None, len(authByteStr), readonly=False)

        # Fill the byte sequence with each byte from the pickled dictionary
        for i,b in enumerate(authByteStr):
            authByteSeq[i] = b

        # Store the buffer size in the class - we need it on retrieval
        self._bufferSize = len(authByteSeq)

        # Write the authorization to the Registry
        if c4d.plugins.WriteRegInfo(self._AUTH_PLUGID, authByteSeq):
            self.logger.info("Wrote Auth to Reg, size: %s", self._bufferSize)
        else:
            self.logger.error("Failed to write Auth, size: %s", self._bufferSize)

        # Store some public-ish parameters in the world container
        tbxPrefsBc = c4d.BaseContainer()
        ### TODO: Use Constants
        tbxPrefsBc.SetLong(20004,self.lastUpdate)
        tbxPrefsBc.SetLong(20005, self.groupID)
        tbxPrefsBc.SetLong(20001, self.memberID)
        tbxPrefsBc.SetString(20002, self.username)
        tbxPrefsBc.SetLong(20003, self._bufferSize)
        if c4d.plugins.SetWorldPluginData(self._AUTH_PLUGID, tbxPrefsBc):
            self.logger.info("Wrote Auth to Plugin Data")

        return True

    def readAuth(self):
        """
        Read the authorization dictionary from the registry
        """

        # Read public-ish parameters from the world container
        tbxPrefsBc = c4d.plugins.GetWorldPluginData(self._AUTH_PLUGID)
        if not tbxPrefsBc:
            return False

        self.lastUpdate = tbxPrefsBc.GetLong(20004)
        self.groupID = tbxPrefsBc.GetLong(20005)
        self.memberID = tbxPrefsBc.GetLong(20001)
        self.username = tbxPrefsBc.GetString(20002)
        self._bufferSize = tbxPrefsBc.GetLong(20003)
        self.logger.info("Read Auth from Plugin Data")

        if self.lastUpdate + self.authExpires < time.time():
            return self.getAuthFromCineversity(memberID=self.memberID)

        # Read the registry to a ByteSeq
        if self._bufferSize > 0:
            authBytes = c4d.plugins.ReadRegInfo(self._AUTH_PLUGID, self._bufferSize)

            if authBytes:
                try:
                    # Convert ByteSeq to str and unpickle
                    self.authDict = pickle.loads(str(authBytes))
                    self.logger.info("Read Auth from Reg")
                except EOFError:
                    # Pickle failed, either the reg key is zeroed or corrupt
                    return False

                return True

        return False

    def clearAuth(self):
        """
        Clears stored authorization info (Logoff)
        """
        self.__init__()
        c4d.plugins.SetWorldPluginData(self._AUTH_PLUGID, c4d.BaseContainer())
        c4d.plugins.WriteRegInfo(self._AUTH_PLUGID, c4d.storage.ByteSeq(None,1))
        self.logger.info("Logged Off User: %s", self.username)


def getXml(url, cacheExpires=0, storeCache=False):
    """
    Downloads and parses an XML file from a url

    :param url:
        The url location of the XML file
    :param cacheExpires:
        Seconds before cache expires and file should be re-fetched
    :param storeCache:
        Whether or not to store a cache file
    :return:
        XML ElementTree
    """

    xml = None
    cachePath = os.path.join(c4dStartupPath,"cvToolbox","toolbox.xml")

    if cacheExpires > 0:
        # Check cache
        bcData = c4d.plugins.GetWorldPluginData(1031620)
        if bcData and bcData.GetLong(20000)+cacheExpires < int(time.time()) and os.path.exists(cachePath):
            with open(cachePath,'r') as xmlFile:
                xml = xmlFile.read()

    if xml is None:
        if url[:4] == "http":
            # Get the file object
            try:
                xml = urllib2.urlopen(url)
                coreLogger.info('Retrieved Resource XML from Cineversity')
            except URLError as e:
                coreLogger.exception('Error retrieving XML', e)
                raise CVToolboxError('Error retrieving XML: %s' % e)
            except Exception:
                import traceback
                coreLogger.exception('GetXML Generic Exception: ' + traceback.format_exc())
                raise CVToolboxError('Error retrieving XML: Generic Exception')
        else:
            if os.path.exists(url):
                xml = url
            else:
                return None

    if storeCache:
        # Cache the XML locally
        with open(cachePath,'w') as xmlFile:
            xmlFile.write(xml)

    # Parse file to ElementTree
    try:
        xmlTree = ET.parse(xml, parser=UnicodeXMLTreeBuilder())
    except Exception:
        coreLogger.exception('Error parsing XML')
        raise CVToolboxError('Error parsing XML')

    coreLogger.info('Parsed Resource XML')

    return xmlTree

def writeResourceXml(root,obj,doc):
    if isinstance(root,str):
        root = ET.Element(root)

    for key,value in obj.items():

        # Some of the obj keys are just for me and shouldn't be logged
        if key=='selected' or \
           key=='avail' or \
           key=='avail_icon' or \
           key=='avail_reason' or \
           key=='idx' or \
           key=='checked' or \
           key=='_download':
            continue
        coreLogger.debug("key=%s,value=%s,dict=%s,list=%s",key,value,isinstance(value,dict),isinstance(value,list))
        el = ET.SubElement(root,str(key))
        if isinstance(value,dict):
            writeResourceXml(el,value,doc)
        elif isinstance(value,list) and len(value)>=1:
            if isinstance(value[0],str) or (not PY3 and isinstance(value[0],unicode)):
                el.text = value[0]
            for i in range(1,len(value)):
                if isinstance(value[i],dict):
                    writeResourceXml(el,value[i],doc)
                else:
                    el = ET.SubElement(root,str(key))
                    el.text = value[i]
        elif isinstance(value,str) or (not PY3 and isinstance(value,unicode)):
            el.text = value.strip()
        else:
            try:
                el.text = str(value).strip()
            except:
                coreLogger.debug('Failed to store XML element: %s=%s[%s]',key,value,type(value))

    return root

def installLogXml(resultsDict,doc=None):
    coreLogger.debug("INSTALL LOG XML DOC: %s",doc)
    if isinstance(doc,ET.ElementTree):
        coreLogger.debug("ElementTree")
        root = doc.getroot()
    elif ET.iselement(doc) or isinstance(doc,type(ET.Element(None))):
        coreLogger.debug("Appending to existing Element")
        root = doc
    else:
        coreLogger.debug("No Doc. Initializing")
        root = ET.Element(str("resources"))
    for i in resultsDict:
        resNode = ET.SubElement(root,str("resource"))
        writeResourceXml(resNode,i,doc)
    return root

def writeXml(elem):
    out = str('<?xml version="1.0" encoding="UTF-8"?>')
    out += str('<resources>')
    for e in elem:
        out += ET.tostring(e).decode('utf-8')
    out += str('</resources>')
    return out

def xmlToTreeDict(xml, filterStr=None):
    """
    Convert ElementTree xml object to a dictionary for use with TreeViewCustomGui
    Simple filtering of string within element text and sorting list by element tag

    This function is not recursive! It only handles single-child trees.

    :param xml:
        XML ElementTree object
    :param filterStr:
        String to search within element text
    :return:
        Python List of Dictionaries, 1 entry per top-level element, dictionary entries for child elements
    """

    xmlTreeList = []
    i = 0

    # Loop through elements in xml ElementTree
    for element in xml.findall("*"):
        filterInclude = False   # init for each element - is filterStr in any element.text
        ret = {'idx': i}        # init ret dict - idx is used by treeview and should always be the index in list
        for child in element:   # loop through each child of top-level elements, add to dict, check filter
            if len(list(child.items())) > 0:
                # if the tag has attributes, we'll store them in a dictionary
                data = dict((k,v) for k,v in list(child.items()))
                if child.text:
                    # if there's text as well, we'll put it in a key called "_text"
                    data['_text'] = child.text.strip()
            else:
                # if there's no attributes, we'll just store text directly
                data = child.text
                if data:
                    data = data.strip()
                else:
                    data = ''
            if child.tag in ret:
                # if this is a duplicate tag, we'll store each tag as a list item
                if type(ret[child.tag]) is type([]):
                    # already a list, just append
                    ret[child.tag].append(data)
                else:
                    # convert to list
                    ret[child.tag] = [ret[child.tag], data]
            else:
                # in most cases, just set the value to the data (either the text or dict of attribs)
                ret[child.tag] = data
            # perform filtering - we only worry about child.text, not attribs
            if filterStr is None or filterStr == "" or (child.text and child.text.lower().find(filterStr.lower()) >= 0):
                filterInclude = True
        if filterStr is None or filterStr == "" or filterInclude is True:      # check filter & append to list
            xmlTreeList.append(ret)
            i+=1                # increment i instead of enumerate since some elements will be filtered

    return xmlTreeList

def parseTime(s,retType='dt'):

    if PY3:
        # PY3 has this method
        d = datetime.datetime.fromisoformat(s)

    else: 
    
        # Perform conversion using strptime - this yields locale-based errors in py3
        fmts = ['%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M', '%Y-%m-%d']
        d = None
        for fmt in fmts:
            try:
                d = datetime.datetime.strptime(s, fmt)
                #coreLogger.debug("Strptime Parse Succeeed with [%s] using format [%s]", s, fmt)
            except ValueError as err:
                #coreLogger.debug("Strptime Parse Failed with [%s] using format [%s] - %s", s, fmt, err)
                pass

    if d:
        if retType == 'int':
            return time.mktime(d.timetuple())
        else:
            return d

def getLastUpdated(xml):
    if xml is None: return None

    latest = datetime.datetime.fromtimestamp(946684800) #min was causing a mktime argError sometimes
    for node in xml.findall('*/updated'):
        thisDate = parseTime(node.text)
        try:
            if latest < thisDate:
                latest = thisDate
        except TypeError:
            coreLogger.debug("GetLastUpdated parse issue: %s", node.text)

    return time.mktime(latest.timetuple())

def getResourceDescription(obj):
    """
    Get the individual description for a single resource, based on HTML template in res folder

    HTML template includes standard Python dictionary substitution formatting, so we use a standard substitution

    :param obj:
        dictionary: keys from xml file
    :return:
        string: html code to display as description
    """

    # Setup path
    if PY3:
        htmlTemplatePath = os.path.join(os.path.dirname(__file__), '..', '..', 'res', 'resourceDescription.html')
    else:
        htmlTemplatePath = os.path.join(unicode(os.path.dirname(__file__),'utf-8'), '..', '..', 'res', 'resourceDescription.html')

    if 'image' in obj and obj['image'] is not None and obj['image'] != '':
        obj['imagehtml'] = '<img class="playlist-thumb" title="More Info" src="%s" alt="" border="0">' % obj['image']
    else:
        obj['imagehtml'] = ''

    if 'status' in obj and obj['status'] == -2:
        obj['statusText'] = '<div class="status">' + \
                            'This plugin was not found in the CV-Toolbox install log, but is currently present in C4D.<br>' + \
                            "This might be because you manually installed it, or because the installLog is incomplete.<br>" + \
                            "If you manually installed, Please remove it and re-install via CV-Toolbox to update automatically.<br>" + \
                            '<a href="https://bitbucket.org/cineversity/cv-toolbox/wiki/Plugin%20Removal">Click here for detailed instructions.</a>' + \
                            "</div>"
    elif 'avail' in obj and obj['avail'] is False:
        reason = ''
        if 'avail_reason' in obj:
            reason = obj['avail_reason']
        if reason == 'premium':
            obj['statusText'] = '<div class="status">' + \
                                'This resource is available for Premium Cineversity members only.<br>' + \
                                'Trial users do not have access to Premium resources.<br>' + \
                                'If you have a coupon, please visit <a href="http://www.cineversity.com">cineversity.com</a> to redeem your coupon.' + \
                                '</div>'
        elif reason == 'protected':
            obj['statusText'] = '<div class="status">' + \
                                'This resource is available only with a special Cineversity coupon.<br>' + \
                                'If you have a coupon, please visit <a href="http://www.cineversity.com">cineversity.com</a> to redeem your coupon.' + \
                                '</div>'
    else:
        obj['statusText'] = ''

    # Open the html template
    with open(htmlTemplatePath) as htmlFile:
        html = htmlFile.read()          # read the template
        return html % obj               # perform dictionary substitution and return

def getInstallPath(resType, allUsers=True, pluginInstallDirectory=None):
    """
    Get the proper install path for a resource, based on the resource type

    :param resType:
        string: type of resource
    :param allUsers:
        bool: if true, install for all users (application directory), otherwise in user's home directory
    :return:
        string: output path
    """

    resType = resType.lower()

    if allUsers:
        # Installing for all users starts with the C4D application path
        if PY3:
            prefixPath = c4d.storage.GeGetStartupPath()
        else:
            prefixPath = unicode(c4d.storage.GeGetStartupPath(),'utf-8')
    else:
        # Otherwise, start with the user home directory
        prefixPath = c4dStartupPath

    # Create path based on resource type
    if resType=='plugin':
        # plugins can be installed in custom paths
        if pluginInstallDirectory is None:
            return os.path.join(prefixPath,'plugins')
        else:
            return pluginInstallDirectory

    # Everything else goes in he prefs-related path
    elif resType=='script':
        return os.path.join(prefixPath,'library','scripts')
    elif resType=='contentlibrary' or resType=='contentpreset':
        return os.path.join(prefixPath,'library','browser')
    elif resType=='character':
        return os.path.join(prefixPath,'library','characters')
    elif resType=='cmotion':
        return os.path.join(prefixPath,'library','cmotion')
    elif resType=='pythonpackage':
        return c4d.storage.GetUserSiteSpecificPath()            # python packages always go here
    elif resType=='xgroup':
        return os.path.join(prefixPath,'library','xgroup')
    elif resType=='xpool':
        return os.path.join(prefixPath,'library','xnode')
    elif resType=='sketchstyle':
        return os.path.join(prefixPath,'library','sketch')
    elif resType=='materialpreview':
        return os.path.join(prefixPath,'library','materialpreview')
    elif resType=='layout':
        return os.path.join(prefixPath,'library','layout')
    elif resType=='schemes':
        # not sure if schemes can go in C4D application path
        # for now, force them into the user path
        return os.path.join(c4dStartupPath,'library', 'schemes')
    else:
        # should never get here, but just in case
        coreLogger.info('Unknown resource type: ', resType)
        return None

def installedTreeDict(filterStr=None):
    pl_list = c4d.plugins.FilterPluginList(c4d.PLUGINTYPE_ANY,True)
    resTypes = ['plugin','script','contentlibrary','cmotion','character','pythonpackage','xgroup','xpool',
                'sketchstyle','materialpreview','layout','schemes']
    extTypes = ['.py','.pyp','.pype','.pypv','.csc','.lib4d','.cdl64','.dylib','.c4d','.l4d']
    for resType in resTypes:
        for allUsers in [True,False]:
            path = getInstallPath(resType,allUsers)
            for root,dirs,files in os.walk(path):
                for f in files:
                    ext = os.path.splitext(f)[1]
                    if ext.lower() in extTypes:
                        if allUsers:
                            allUsersTxt = "Global"
                        else:
                            allUsersTxt = "User"
                        coreLogger.debug("Found %s :: %s :: %s :: %s" % (resType,allUsersTxt,os.path.basename(root),f))

def checkPathWrite(path):

    # We'll try os.access first, but os.access always returns true in windows
    if not os.access(path, os.W_OK | os.X_OK):
        print("OS.ACCESS Failed")
        return False
    
    # So if os.access succeeds we'll double-check with a pythonic test
    filetest = os.path.join(path,'tbxtest')
    try:
        with open(filetest, 'w') as f:
            coreLogger.debug('Write Test succeeded for path %s', filetest)
        os.remove(filetest)
        return True
    except IOError as x:
        print(x.errorText)
        if x.errno == errno.EACCES:
            coreLogger.debug('Write test failed for %s: Access Permission Failed', filetest)
        return False
    
class CVBoxDownloadInstall(c4d.threading.C4DThread):
    """
    Custom thread to perform downloading and installation of CVToolbox Resources
    """

    _listToInstall = None           # internal list of dict resources to install
    _progress1 = 0                  # overall progress
    _progress2 = 0                  # individual progress
    _statusText = ""                # status text
    _allUsers = False               # install global or local
    _cachePath = None               # path to store cached download files
    _chunkSize = 8192               # chunk size to download each loop

    def __init__(self, listToInstall, allUsers=False, memberID = 0, groupID = 0, pluginInstallPath = None):
        """
        Initialize class

        :param listToInstall:
            list of resources to install - each resource is a dict with keys from XML
        """
        self._listToInstall = listToInstall
        self._allUsers = allUsers
        self._pluginInstallPath = pluginInstallPath
        self._memberID = memberID
        self._groupID = groupID
        self._results = []
        self._failCount = 0
        self._fileCleanup = []

        self.errorText = ""
        self.errorAnswer = None

        self.logger = logging.getLogger('cvToolbox.core.install')

    def Main(self):
        """
        Main thread operation - setup, master loop
        """

        # Set up cache path, create directory if necessary
        self._cachePath = str(os.path.join(c4dStartupPath,"cvToolbox","cache"))
        if not os.path.isdir(self._cachePath):
            os.makedirs(self._cachePath)

        # Master loop - loop through each resource to install
        for i, obj in enumerate(self._listToInstall):

            # Check if the thread needs to end
            if self.TestBreak():
                self.logger.info("Download & Install interrupted in master loop")
                break

            # Download the file
            self.logger.debug("Downloading %s", obj['name'])
            try:
                f = self.downloadFile(obj, i)
                if f:
                    self._statusText = "Installing " + obj['name']
                    self.logger.debug("Installing %s", obj['name'])
                    self.installFile(obj, f, i)
                    #os.remove(f)
            except CVToolboxError:
                # on error, we just want to continue looping through each item to install
                self._failCount += 1
                continue

        # Delete the symbolcache
        try:
            symbolcache = os.path.join(c4dStartupPath,'prefs','symbolcache')
            os.remove(symbolcache)
        except OSError:
            pass

        # Store a cleanup log
        if len(self._fileCleanup) > 0:
            try:
                cleanLogFn = os.path.join(c4dStartupPath,"cvToolbox","cleanup.log")
                with io.open(cleanLogFn,'a',encoding='utf-8') as cleanLog:
                    for item in self._fileCleanup:
                        cleanLog.write('%s\n' % item)
            except IOError:
                raise CVToolboxError("Couldn't write cleanup log")

        # Tell the dialog we're done
        ### TODO: Use Constant instead of hard-coded value
        c4d.SpecialEventAdd(1031623)

    def getProgress(self):
        """
        Return current thread progress. Called from outside thread since GUI ops can't be done in thread.

        :return:
            tuple: overall progress, individual progress
        """
        return self._progress1, self._progress2, self._statusText

    def guiMessageDialog(self, msg, btnType):
        self.errorAnswer = None
        self.errorText = msg
        c4d.SpecialEventAdd(1031624)
        while self.errorAnswer is None:
            if self.TestBreak():
                self.logger.info("User interrupted with Message open")
                # Remove the partially-downloaded file
                break
            time.sleep(.5)
        return self.errorAnswer

    def getResults(self):
        """
        Return the results of the installation

        :return:
            tuple: number of failures, list of dictionaries for failures
        """
        return self._failCount, self._results

    def downloadFile(self, obj, index):
        """
        Download an individual file

        :param obj:
            dictionary: keys from xml for single object/resource
        :param masterPct:
            float: current overall percentage of completion (curitem/itemcount)

        :raises:
            CVToolboxError

        :return:
            str: file path of downloaded file
        """

        downloadDict = {}
        obj['_download'] = []

        # Get the download url based on the C4D version
        if type(obj['download']) is type([]):
            # Object has multiple download links, we need to pick the right one
            c4dVer = c4d.GetC4DVersion()        # Get the C4D Version
            # Loop through download links, checking which has a version between first and last
            for i in obj['download']:
                if 'last' in i and int(i['last']) == 0: i['last'] = 99999
                #coreLogger.debug("First: %s, Last: %s",i['first'],i['last'])
                if 'first' in i and 'last' in i and (int(i['first']) <= c4dVer < int(i['last'])):
                    #self.logger.debug("Found compatible version for C4D Version: %s", c4dVer)
                    obj['_download'].append(i)
            if not obj['_download']:
                # If no appropriate download is found, we'll log an error and assume the first
                self.logger.error("Compatible version not found for C4D Version: %s", c4dVer)
                obj['_download'] = obj['download']
        else:
            obj['_download'].append(obj['download'])

        for i,d in enumerate(obj['_download']):

            # Get Download URL from obj, might be in _text dictionary key or direct value
            if type(d) is type({}):
                downloadUrl = d['_text']
            else:
                downloadUrl = d
                obj['_download'][i] = {'_text':downloadUrl}

            # Skip this download if we've already done it
            if downloadUrl in downloadDict:
                obj['_download'][i]['filePath'] = downloadDict[downloadUrl]
                continue

            # Generate post data for Cineversity download requests
            postValues = {}
            if obj['access'] != 'external':
                now = datetime.datetime.utcnow()
                postValues['memberID'] = self._memberID
                postValues['groupID'] = self._groupID
                postValues['hash'] = hashlib.sha1(str(self._memberID).encode('utf-8') + \
                    str(self._groupID).encode('utf-8') + \
                    str(now.year).encode('utf-8') + \
                    str(now.month).encode('utf-8') + \
                    str(now.day).encode('utf-8')).hexdigest()
                self.logger.debug('Downloading with CV Auth Member ID: %s, Group ID: %s',self._memberID,self._groupID)
                #print "%s&memberID=%s&groupID=%s&hash=%s" % (downloadUrl,self._memberID,self._groupID,postValues['hash'])
            postData = urlencode(postValues)

            # Open the remote file
            try:
                response = urllib2.urlopen("%s&memberID=%s&groupID=%s&hash=%s" % (downloadUrl,self._memberID,self._groupID,postValues['hash']))
            except URLError as e:
                self._results.append({'obj': obj, 'success': False, 'msg': 'Failed to download files - check internet connection', 'err': e})
                self.logger.error('Failed to download files: %s, %s', obj['download'], e)
                raise CVToolboxError('Error downloading file %s: %s', (obj['download'], e))

            # Determine the file size from the HTTP response header
            # Determine the filename from content disposition header
            if PY3:
                self.logger.debug('Response Headers:\n%s',response.getheaders())
                totalSize = response.getheader('Content-Length')
                contentType = response.getheader('Content-Type')
                content = response.getheader('Content-Disposition')
            else:
                self.logger.debug('Response Headers:\n%s',response.info())
                totalSize = response.info().getheader('Content-Length')
                contentType = response.info().getheader('Content-Type')
                content = response.info().getheader('Content-Disposition')

            if totalSize:
                totalSize = int(totalSize.strip())
            else:
                # No size given in header / header not present
                totalSize = -1      # Setting totalSize to negative will cause status bar to spin
                self.logger.info("Unable to determine size of %s", obj['name'])

            bytesRead = 0

            if contentType is None or (contentType != "application/zip" and contentType != "application/octet-stream"):
                if contentType == 'text/plain' or contentType == 'text/html':
                    contentErr = response.read(50)
                else:
                    contentErr = 'specific reason unknown'
                    self.logger.debug("Content Type: %s",contentType)
                    self.logger.debug("Response: %s",response.read(100))
                # Most common reason for this will be with CV file links if the authorization isn't correct
                self._results.append({'obj': obj, 'success': False, 'msg': 'Downloaded file is not a valid plugin/resource: ' + contentErr, 'err': 'CV Auth Error or File Not Found'})
                self.logger.info("Error downloading %s:: CV Auth Error or File Not Found", obj['name'])
                raise CVToolboxError("Error downloading %s:: CV Auth Error or File Not Found", obj['name'])

            # Sometimes filename won't be provided - currently falling back to id parameter
            filename = obj['id'] + "_" +str(i)
            
            if content is not None:
                # Assuming we have Content Disposition, parse the filename from the header
                filenameMatch = re.findall('filename=(\S+)', content)
                if filenameMatch:
                    filename = filenameMatch[0].replace('"','')

            if not PY3 and not isinstance(filename,unicode):
                filename = unicode(filename,'utf-8')

            # Define complete path, joining cachepath and filename, and open destination file
            filePath = os.path.join(self._cachePath,filename)
            f = open(filePath,'wb')

            while True:
                # Check if we need to break the thread and quit downloading
                if self.TestBreak():
                    self.logger.info("Download Interrupted by User")
                    f.close()
                    # Remove the partially-downloaded file
                    os.remove(filePath)
                    return None

                # Read a chunk of the response
                chunk = response.read(self._chunkSize)
                if not chunk:
                    # nothing read. end of file.
                    break

                # Write the chunk
                f.write(chunk)

                # Update the total bytes read and resulting class progress properties
                bytesRead += len(chunk)

                pctRead = float(bytesRead) / totalSize
                if totalSize > 0:
                    self._statusText = "Downloading " + obj['name'] + " {0:.0%}".format(pctRead)
                else:
                    self._statusText = "Downloading " + obj['name']
                masterPct = float(index+1)/len(self._listToInstall)
                self._progress1 = (float(index+1)/len(self._listToInstall)) + (pctRead*0.5*(1.0/len(self._listToInstall)))
                self._progress2 = pctRead * 0.5

                # Tell the dialog to update the progress bars
                ### TODO: Use Constant instead of hard-coded value
                c4d.SpecialEventAdd(1031622)

            self.logger.debug("File Path: %s",filePath)
            self.logger.debug("_D: %s", obj['_download'])
            obj['_download'][i]['filePath'] = filePath
            self.logger.debug("D: %s", d)
            downloadDict[downloadUrl] = filePath

        return downloadDict

    def installFile(self, obj, filePath, index):
        """
        Extract resource file to proper path based on type

        :param obj:
            dictionary: keys from xml for single object/resource
        :param filePath:
            string: path to zip file for extration
        :param masterPct:
            float: current overall percentage of completion (curitem/itemcount)

        :raises:
            CVToolboxError

        :return:
            boolean: success
        """

        installed_files = []

        self.logger.debug("Download dict: %s", obj['_download'])
        self.logger.debug("Obj dict: %s", obj)

        for d in obj['_download']:

            if not 'filePath' in d: continue

            filePath = d['filePath']
            if 'type' in d:
                outputPath = getInstallPath(d['type'], self._allUsers, self._pluginInstallPath)
            else:
                outputPath = getInstallPath(obj['type'], self._allUsers, self._pluginInstallPath)

            ###TODO: Modify output path if obj['pluginid'] is installed
            if obj['type'] == 'Plugin' and 'pluginid' in obj:                  
                #Ideally, we'd use FindPlugin and GetFilename to get the path of the installed plugin
                #But GetFilename returns just the file, not the directory
                #plug = c4d.plugins.FindPlugin(int(obj['pluginid']))
                #if plug:
                #     outputPath = os.path.dirname(plug.GetFilename())

                #Instead for now we'll grab the path from the install log
                if 'installPath' in obj:
                    outputPath = obj['installPath']

            if not PY3 and not isinstance(outputPath,unicode):
                outputPath = unicode(outputPath,'utf-8')

            self.logger.debug("Extracting from %s", filePath)
            self.logger.debug("Installing to %s", outputPath)

            zf = zipfile.ZipFile(filePath)

            uncompress_size = sum((file.file_size for file in zf.infolist()))

            extracted_size = 0

            # store dirFilter for later use
            dirFilter = None
            addDir = ""
            baseDirs = []
            if type(d) is type({}):
                if 'dir' in d:
                    # ensure that the dir filter doesn't have a trailing slash - it would break base directory logic
                    dirFilter = d['dir'].rstrip("/")
                if 'adddir' in d:
                    # ensure that addDir doesn't have leading or trailing slashes
                    addDir = d['adddir'].strip("\\ /")

            # detect the base directory
            if addDir is None or addDir == '':
                for name in zf.namelist():
                    if name.startswith("__") or name.startswith("._"):
                        # ignore hidden folders
                        continue
                    if dirFilter is None or name.startswith(dirFilter):
                        if dirFilter is not None:
                            name = name[len(dirFilter):]
                        if name.endswith("/"):
                            namesplit = name.strip('/').split('/')
                            if len(namesplit) == 1 and namesplit[0] != '':
                                baseDirs.append(namesplit[0])
            else:
                baseDirs.append(addDir)

            self.logger.debug("Base Directories: %s", baseDirs)

            # check if the destination folder already exists
            if type(d) is type({}) and 'replaceall' in d:
                if d['replaceall'].lower() == 'true':
                    for f in baseDirs:

                        if not PY3 and not isinstance(f,unicode):
                            f = unicode(f,'utf-8')

                        # check the direct output path
                        if os.path.exists(os.path.join(outputPath, f)):
                            # the same folder we're trying to extract already exists, warn the user
                            ans = self.guiMessageDialog("It looks like %s is already installed. Is it ok to remove the installed version?" % obj['name'], c4d.GEMB_YESNO)
                            if ans == c4d.GEMB_R_YES:
                                self.logger.info("Archiving folder %s", os.path.join(outputPath, f))
                                self.archiveExistingFile(os.path.join(outputPath, f))
                            else:
                                self._failCount += 1
                                self._results.append({'obj': obj, 'success': False, 'msg': 'Existing plugin detected. User denied overwrite.', 'err': 'User denied overwrite of pre-existing folder'})
                                self.logger.info("User denied overwrite of pre-existing folder")
                                return False
                        # also check the related path
                        elif os.path.exists(os.path.join(getInstallPath(obj['type'], not self._allUsers),f)):
                            ans = self.guiMessageDialog("It looks like %s is already installed in the global plugins directory.\nPlease remove manually to avoid plugin conflicts." % obj['name'], c4d.GEMB_OK)

            for f in zf.infolist():
               # Check if we need to break the thread and quit downloading
                if self.TestBreak():
                    self.logger.info("Install interrupted by user")
                    # Remove the partially-downloaded file
                    break

                if f.filename.startswith("__") or f.filename.startswith("._"):
                    # don't extract hidden files and folders created by MacOS
                    continue

                # for objects with a 'dir' dictionary item, only extract files/folders inside dir
                if dirFilter:
                    self.logger.debug("DIR FILTER: %s, %s",f.filename, dirFilter)
                    if f.filename[:len(dirFilter)] == dirFilter:
                        # file/folder is inside the filtered directory
                        fname = os.path.split(f.filename)   # get filename parts, used to determine if it's a folder or file
                        zipPath = f.filename[len(dirFilter):]   # get the folder hierarchy starting from the filtered dir
                        self.logger.debug("Path: %s, %s, %s",outputPath,addDir,zipPath)
                        fullPath = os.path.normpath(os.path.join(outputPath,addDir) + zipPath)   # append folder hierarchy to existing output path
                        modDate = time.mktime(f.date_time + (0,0,-1))
                        if fname[1]:
                            # it's a file
                            try:
                                outfile = open(fullPath, 'wb')          # create the destination file
                                outfile.write(zf.read(f))               # write the file
                                outfile.close()                         # close the file
                                os.utime(fullPath, (modDate,modDate))
                                os.chmod(fullPath, 0o755)
                                installed_files.append(fullPath)
                            except:
                                try:
                                    self.archiveExistingFile(fullPath)  # if the file already exists and is locked, archive the existing file
                                    outfile = open(fullPath, 'wb')      # create the destination file
                                    outfile.write(zf.read(f))           # write the file
                                    outfile.close()                     # close the file
                                    os.utime(fullPath, (modDate,modDate))
                                    os.chmod(fullPath, 0o755)
                                    installed_files.append(fullPath)
                                except IOError as e:
                                    self.logger.exception("Couldn't write file: %s, %s", f.filename, e)
                                    raise CVToolboxError("Couldn't write file: %s, %s", f.filename, e)
                                except OSError as e:
                                    self.logger.exception("Couldn't write file: %s, %s", f.filename, e)
                                    raise CVToolboxError("Couldn't write file: %s, %s", f.filename, e)
                        else:
                            # it's a directory
                            try:
                                if not os.path.exists(fullPath):
                                    os.mkdir(fullPath)
                            except OSError as e:
                                self.logger.exception("Couldn't create directory: %s, %s", fullPath, e)
                                raise CVToolboxError("Couldn't create directory: %s, %s", fullPath, e)
                else:
                    try:
                        zf.extract(f, os.path.join(outputPath,addDir))
                        os.chmod(os.path.join(outputPath,addDir,f.filename),0o755)
                        installed_files.append(os.path.join(outputPath,addDir,f.filename))
                        self.logger.debug("Installed Files: %s",installed_files)
                    except IOError:
                        try:
                            self.archiveExistingFile(os.path.join(outputPath, f.filename))
                            zf.extract(f, outputPath)
                            os.chmod(os.path.join(outputPath,f.filename),0o755)
                            installed_files.append(os.path.join(outputPath, f.filename))
                        except IOError as e:
                            self.logger.exception("Couldn't write file: %s, %s", f.filename, e)
                            raise CVToolboxError("Couldn't write file: %s, %s" % (f.filename, e))
                        except OSError as e:
                            self.logger.exception("Couldn't write file: %s, %s", f.filename, e)
                            raise CVToolboxError("Couldn't write file: %s, %s" % (f.filename, e))

                extracted_size += f.file_size

                pctExtracted = float(extracted_size) / uncompress_size
                self._statusText = "Installing " + obj['name'] + " {0:.0%}".format(pctExtracted)
                self._progress1 = (float(index+1)/len(self._listToInstall)) + ((0.5+pctExtracted*0.5)*(1.0/len(self._listToInstall)))
                self._progress2 = 0.5 + (pctExtracted * 0.5)

                # Tell the dialog to update the progress bars
                ### TODO: Use Constant instead of hard-coded value
                c4d.SpecialEventAdd(1031622)

            # Success
            zf.close()              # close the zipfile

        # Once all installs are done, remove cache files
        # Needs to be done after all installs are processed, since some install lines may share a single download file
        for d in obj['_download']:
            if 'filePath' in d:
                try:
                    os.remove(d['filePath'])
                except OSError:
                    pass

        obj['installed'] = datetime.datetime.now()
        obj['installPath'] = outputPath
        obj['installFiles'] = installed_files
        self._results.append({'obj': obj, 'success': True, 'msg': 'Installation Successful'})
        return True

    def archiveExistingFile(self, f, oldRoot="", newRoot=""):
        """
        If filepath provided exists, rename .old

        :param f:
            filepath string
        :return:
            new filepath
        """
        if not os.path.exists(f): return None

        timestamp = int(time.mktime(datetime.datetime.now().timetuple()))
        newName = os.path.join(f+".old"+str(timestamp))

        if os.path.isdir(f):
            for root, dirs, files in os.walk(f):
                for name in files:
                    try:
                        self.archiveExistingFile(os.path.join(root,name),f,newName)
                    except IOError:
                        raise CVToolboxError("Couldn't Rename Directory for Archival: %s" % f)
                    except OSError:
                        raise CVToolboxError("Couldn't Rename Directory for Archival: %s" % f)
            os.rename(f,newName)
            self._fileCleanup.append(newName)
            self.logger.debug('Renamed folder for overwrite: %s', newName)
        else:
            try:
                os.rename(f,newName)
                if oldRoot != "" and newRoot != "":
                    newName = newRoot + newName[len(oldRoot):]
                self._fileCleanup.append(newName)
                self.logger.debug('Renamed file for overwrite: %s', newName)
                return newName
            except IOError as e:
                self.logger.exception("Couldn't Rename file for overwrite: %s", newName)
                raise CVToolboxError("Couldn't Rename file for overwrite: %s" % newName)